@@ -1,5 +1,5 @@
 //
-// $Id: dnstracer.c,v 1.37 2002/07/30 11:26:04 mavetju Exp $
+// $Id: dnstracer.c,v 1.40 2003/02/14 09:58:54 mavetju Exp $
 //
 
 //
@@ -64,6 +64,7 @@
 #define DEFAULT_CACHING	1
 #define DEFAULT_OVERVIEW	0
 #define DEFAULT_QUERYTYPE	ns_t_a
+#define DEFAULT_NOIPV6	0
 
 int	verbose=0;
 int	global_overview=DEFAULT_OVERVIEW;
@@ -71,6 +72,8 @@ int	global_retries=DEFAULT_RETRIES;
 int	global_timeout=DEFAULT_TIMEOUT;
 int	global_caching=DEFAULT_CACHING;
 int	global_querytype=DEFAULT_QUERYTYPE;
+int	global_noipv6=DEFAULT_NOIPV6;
+char *	global_source_address=NULL;
 
 /*****************************************************************************/
 
@@ -115,6 +118,7 @@ struct dnssession {
     char *			recv_pkt_additional;	// - first additional RR
 
     int				socket;
+    int				ipv6;
 };
 
 struct dnsheader {
@@ -599,7 +603,7 @@ void dump_rr(struct dnsrr *rr,struct dnssession *session) {
 				    get_resource(rr->type,session,rr->data,0));
 }
 
-void dump_data( struct sockaddr_in *dest,struct dnssession *session) {
+void dump_data(struct sockaddr_in *dest4,struct sockaddr_in6 *dest6,struct dnssession *session) {
     struct dnsrr *answerrr;
     struct dnsrr *authorityrr;
     struct dnsrr *additionalrr;
@@ -610,9 +614,13 @@ void dump_data( struct sockaddr_in *dest,struct dnssession *session) {
     // Dumps the output of session on the screen.
     //
 
-    if (dest!=NULL) {
+    if (dest4!=NULL) {
 	printf("IP HEADER\n");
-	printf("- Destination address:  %s\n",inet_ntoa(dest->sin_addr));
+	printf("- Destination address:  %s\n",inet_ntoa(dest4->sin_addr));
+    }
+    if (dest6!=NULL) {
+	printf("IP HEADER\n");
+	printf("- Destination address:  %s\n","XXX");
     }
 
     if (session->send_header!=NULL && session->recv_header==NULL) {
@@ -702,14 +710,42 @@ unsigned char *create_packet(struct dnssession *session) {
 // Network routines
 //
 
-int create_socket(void) {
+int create_socket(int PF) {
     int	s;
 
-    if ((s=socket(PF_INET,SOCK_DGRAM,0))<0) {
+    if ((s=socket(PF,SOCK_DGRAM,0))<0) {
 	perror("create_socket/socket");
+	printf("If this is an IPv6 problem, run configure with --disable-ipv6\n");
 	exit(1);
     }
 
+    if (global_source_address!=NULL) {
+	struct addrinfo hints,*src_res;
+	int error;
+
+	// thanks to the src/usr.bin/telnet/commands.c of FreeBSD 4.7!
+        memset(&hints, 0, sizeof(hints));
+        hints.ai_flags = AI_NUMERICHOST;
+        hints.ai_family = PF_INET;
+        hints.ai_socktype = SOCK_DGRAM;
+        error = getaddrinfo(global_source_address, 0, &hints, &src_res);
+        if (error == EAI_NODATA) {
+	    hints.ai_flags = 0;
+	    error = getaddrinfo(global_source_address, 0, &hints, &src_res);
+        }
+        if (error != 0) {
+	    perror(global_source_address);
+	    if (error == EAI_SYSTEM)
+		perror(global_source_address);
+	    exit(1);
+        }
+
+	if (bind(s,src_res->ai_addr,src_res->ai_addrlen)<0) {
+	    perror("create_socket/bind");
+	    exit(1);
+	}
+    }
+
     return s;
 }
 
@@ -717,26 +753,46 @@ int send_data(char *server,struct dnssession *session) {
     int		cc;
     char *	pkt;
     int		len;
-    struct sockaddr_in dest;
+    struct sockaddr_in dest4;
+#ifndef NOIPV6
+    struct sockaddr_in6 dest6;
+#endif
 
     len=sizeof(struct dnsheader)+session->send_question->querylength+4;
 
-    memset(&dest,0,sizeof(struct sockaddr_in));
-    dest.sin_family=AF_INET;
-    dest.sin_port=htons(53);
-    dest.sin_addr.s_addr=inet_addr(server);
+#ifndef NOIPV6
+    if (session->ipv6) {
+	memset(&dest6,0,sizeof(struct sockaddr_in6));
+	dest6.sin6_family=AF_INET6;
+	dest6.sin6_port=htons(53);
+	inet_pton(AF_INET6,server,&dest6.sin6_addr);
+
+	dump_data(NULL,&dest6,session);
+
+    } else
+#endif
+    {
+	memset(&dest4,0,sizeof(struct sockaddr_in));
+	dest4.sin_family=AF_INET;
+	dest4.sin_port=htons(53);
+	dest4.sin_addr.s_addr=inet_addr(server);
+
+	dump_data(&dest4,NULL,session);
+    }
 
-    dump_data(&dest,session);
     pkt=create_packet(session);
-
     if ((cc=sendto(session->socket,pkt,len,0,
-		(struct sockaddr *)&dest,sizeof(struct sockaddr_in)))==-1) {
+#ifndef NOIPV6
+	    session->ipv6?(struct sockaddr *)&dest6:(struct sockaddr *)&dest4,
+	    session->ipv6?sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)
+#else
+	    (struct sockaddr *)&dest4,sizeof(struct sockaddr_in)
+#endif
+	))==-1) {
 	perror("send_data/sendto");
-//	exit(1);
     }
 
     free(pkt);
-
     return cc;
 }
 
@@ -769,7 +825,7 @@ int receive_data(struct dnssession *session) {
     session->recv_pkt=(char *)calloc(1,len);
     memcpy(session->recv_pkt,buffer,len);
     extract_data(session);
-    dump_data(NULL,session);
+    dump_data(NULL,NULL,session);
     return 0;
 }
 
@@ -958,7 +1014,8 @@ int is_busy(char *server) {
 //
 
 int create_session(char *host,
-		    char *server_ip,char *server_name,char *server_authfor,
+		    char *server_ip,int ipv6,
+		    char *server_name,char *server_authfor,
 		    int depth,char *prefix,int last) {
     struct dnssession * session;
     struct dnsrr *	rrauth;
@@ -984,6 +1041,17 @@ int create_session(char *host,
 	return 1;
     }
 
+    if (global_noipv6 && ipv6) {
+	printf("(%s) Not queried",server_ip);
+	return 1;
+    }
+#ifdef NOIPV6
+    if (ipv6) {
+	printf("(%s) Not queried",server_ip);
+	return 1;
+    }
+#endif
+
     printf("(%s) ",server_ip);
     fflush(stdout);
 
@@ -1016,7 +1084,8 @@ int create_session(char *host,
     // don't have to worry about packets received from previous sessions.
     //
     session=(struct dnssession *)calloc(1,sizeof(struct dnssession));
-    session->socket=create_socket();
+    session->socket=create_socket(ipv6?AF_INET6:AF_INET);
+    session->ipv6=ipv6;
     session->server=strdup(server_ip);
     session->host=strdup(host);
     create_header(session);
@@ -1117,7 +1186,6 @@ int create_session(char *host,
 	if (strcmp(rrauth->domainname_string,".") &&
 		strcasecmp(rrauth->domainname_string,
 		    host+strlen(host)-strlen(rrauth->domainname_string))!=0) {
-	    printf("blaat '%s' '%s'\n",host,rrauth->domainname_string);
 	    rrauth=rrauth->next;
 	    continue;
 	}
@@ -1165,27 +1233,21 @@ int create_session(char *host,
 		strcpy(nextserver_name,printablename(rradd->domainname,1));
 		strcpy(nextserver_ip,rradd->data_string);
 		retval+=create_session(host,
-			    nextserver_ip,nextserver_name,rrauth->domainname_string,
+			    nextserver_ip,rradd->type==ns_t_aaaa,
+			    nextserver_name,rrauth->domainname_string,
 			    depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
 	    } else {
-		struct hostent *h;
+		int ip,ipfound=0;
 
 		strcpy(nextserver_name,rrauth->data_string);
 
-		if ((h=gethostbyname(nextserver_name))==NULL) {
-		    //
-		    // No IP address was found for this hostname.
-		    // Just call the function and let them print
-		    // an error.
-		    //
-		    printf("\n");
-		    nextserver_ip[0]=0;
-		    retval+=create_session(host,
-			nextserver_ip,nextserver_name,rrauth->domainname_string,
-			depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
-		} else {
+		for (ip=0;ip<2;ip++) {
 		    int count,i;
-		    char **addr_list;
+		    struct hostent *h;
+		    char **addr_list=NULL;
+
+		    h=gethostbyname2(nextserver_name,ip==0?AF_INET:AF_INET6);
+		    if (h==NULL) continue;
 
 		    //
 		    // One or more IP address were found. Go through all
@@ -1194,7 +1256,6 @@ int create_session(char *host,
 		    //
 		    count=0;
 		    while (h->h_addr_list[count]!=NULL) count++;
-
 		    addr_list=(char **)calloc(count,sizeof(char *));
 		    for (i=0;i<count;i++) {
 			addr_list[i]=(char *)calloc(1,h->h_length);
@@ -1202,18 +1263,43 @@ int create_session(char *host,
 		    }
 
 		    for (i=0;i<count;i++) {
+			if (ip==0) {
+			    unsigned char *s=addr_list[i];
+			    sprintf(nextserver_ip,
+				"%hu.%hu.%hu.%hu",s[0],s[1],s[2],s[3]);
+			    ipv6=0;
+			} else {
+			    unsigned char *s=addr_list[i];
+			    sprintf(nextserver_ip,
+				"%02hx%02hx:%02hx%02hx:%02hx%02hx:%02hx%02hx:"
+				"%02hx%02hx:%02hx%02hx:%02hx%02hx:%02hx%02hx",
+				s[ 0],s[ 1],s[ 2],s[ 3],s[ 4],s[ 5],s[ 6],s[ 7],
+				s[ 8],s[ 9],s[10],s[11],s[12],s[13],s[14],s[15]);
+			    ipv6=1;
+			}
 			printf("\n");
-			sprintf(nextserver_ip,"%hu.%hu.%hu.%hu",
-			    (unsigned char)addr_list[i][0],
-			    (unsigned char)addr_list[i][1],
-			    (unsigned char)addr_list[i][2],
-			    (unsigned char)addr_list[i][3]);
 
 			retval+=create_session(host,
-			    nextserver_ip,nextserver_name,rrauth->domainname_string,
+			    nextserver_ip,ip==1,
+			    nextserver_name,rrauth->domainname_string,
 			    depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
+			ipfound++;
 		    }
 		}
+
+		if (ipfound==0) {
+		    //
+		    // No IP address was found for this hostname.
+		    // Just call the function and let them print
+		    // an error.
+		    //
+		    printf("\n");
+		    nextserver_ip[0]=0;
+		    retval+=create_session(host,
+			nextserver_ip,0,
+			nextserver_name,rrauth->domainname_string,
+			depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
+		}
 	    }
 
 	    //
@@ -1247,7 +1333,7 @@ int create_session(char *host,
 
 void usage(void) {
     fprintf(stderr,
-	"DNSTRACER version 1.5 - (c) Edwin Groothuis - http://www.mavetju.org\n"
+	"DNSTRACER version 1.7 - (c) Edwin Groothuis - http://www.mavetju.org\n"
 	"Usage: dnstracer [options] [host]\n"
 	"\t-c: disable local caching, default enabled\n"
 	"\t-o: enable overview of received answers, default disabled\n"
@@ -1257,6 +1343,8 @@ void usage(void) {
 	"\t             If . is specified, A.ROOT-SERVERS.NET will be used.\n"
 	"\t-t <timeout>: timeout for DNS requests, default 15 seconds\n"
 	"\t-v: verbose\n"
+	"\t-4: don't query IPv6 servers\n"
+	"\t-S <ip address>: use this source address.\n"
     );
     exit(1);
 }
@@ -1277,10 +1365,11 @@ int wsockinit(void) {
 int main(int argc,char **argv) {
     int		ch;
     char *	server_name="127.0.0.1";
-    char *	server_ip="127.0.0.1";
+    char *	server_ip="0000:0000:0000:0000:0000:0000:0000:0000";
     char	ipaddress[NS_MAXDNAME];
     char	argv0[NS_MAXDNAME];
     int		server_root=0;
+    int		ipv6=0;
 
 #ifndef WIN32
     //
@@ -1301,8 +1390,12 @@ int main(int argc,char **argv) {
     wsockinit();
 #endif
 
-    while ((ch=getopt(argc,argv,"coq:r:s:t:v"))!=-1) {
+    while ((ch=getopt(argc,argv,"4coq:r:S:s:t:v"))!=-1) {
 	switch (ch) {
+	case '4':
+	    global_noipv6=1;
+	    break;
+
 	case 'c':
 	    global_caching=0;
 	    break;
@@ -1343,6 +1436,10 @@ int main(int argc,char **argv) {
 	    }
 	    break;
 
+	case 'S':
+	    global_source_address=optarg;
+	    break;
+
 	case 's':
 	    server_name=optarg;
 	    if (strcmp(server_name,".")==0) {
@@ -1383,18 +1480,27 @@ int main(int argc,char **argv) {
 
     {
 	struct hostent *h;
-	if ((h=gethostbyname(server_name))==NULL) {
+	if (((h=gethostbyname2(server_name,AF_INET6))==NULL) &&
+	    ((h=gethostbyname2(server_name,AF_INET))==NULL)) {
 	    fprintf(stderr,"Cannot find IP address for %s\n",server_name);
 	    return 1;
 	}
-	sprintf(ipaddress,"%hu.%hu.%hu.%hu",
-	    (unsigned char)h->h_addr_list[0][0],
-	    (unsigned char)h->h_addr_list[0][1],
-	    (unsigned char)h->h_addr_list[0][2],
-	    (unsigned char)h->h_addr_list[0][3]);
+	if (h->h_addrtype==AF_INET) {
+	    unsigned char *s=h->h_addr_list[0];
+	    sprintf(ipaddress,"%hu.%hu.%hu.%hu",s[0],s[1],s[2],s[3]);
+	    ipv6=0;
+	} else {
+	    unsigned char *s=h->h_addr_list[0];
+	    sprintf(ipaddress,
+		"%02hx%02hx:%02hx%02hx:%02hx%02hx:%02hx%02hx:"
+		"%02hx%02hx:%02hx%02hx:%02hx%02hx:%02hx%02hx",
+		s[ 0],s[ 1],s[ 2],s[ 3],s[ 4],s[ 5],s[ 6],s[ 7],
+		s[ 8],s[ 9],s[10],s[11],s[12],s[13],s[14],s[15]);
+	    ipv6=1;
+	}
     }
 
-    create_session(argv0,ipaddress,server_name,server_root==0?NULL:".",0,"",1);
+    create_session(argv0,ipaddress,ipv6,server_name,server_root==0?NULL:".",0,"",1);
 
     printf("\n");
 
